@@ -1,24 +1,39 @@
 """Implementation of vpype's data model
 """
-import logging
 import math
-from typing import Callable, Dict, Iterable, Iterator, List, Optional, Tuple, Union
+import pathlib
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    Iterator,
+    List,
+    Optional,
+    Set,
+    Tuple,
+    Union,
+    cast,
+)
 
 import numpy as np
 from shapely.geometry import LinearRing, LineString, MultiLineString
 
 from .geometry import crop, reloop
 from .line_index import LineIndex
+from .metadata import (
+    METADATA_FIELD_PAGE_SIZE,
+    METADATA_FIELD_SOURCE_LIST,
+    METADATA_SYSTEM_FIELD_TYPES,
+)
 
-# REMINDER: anything added here must be added to docs/api.rst
 __all__ = [
     "LineCollection",
     "Document",
     "LineLike",
     "LineCollectionLike",
     "as_vector",
-    # deprecated:
-    "VectorData",
+    "_MetadataMixin",  # for documentation
 ]
 
 LineLike = Union[LineString, LinearRing, Iterable[complex]]
@@ -35,8 +50,63 @@ def as_vector(a: np.ndarray):
     return a.view(dtype=float).reshape(len(a), 2)
 
 
+class _MetadataMixin:
+    def __init__(self, metadata: Optional[Dict[str, Any]] = None):
+        self._metadata: Dict[str, Any] = metadata.copy() if metadata else {}
+
+    @property
+    def metadata(self):
+        """Returns the collection's metadata.
+
+        Returns:
+            metadata
+        """
+        return self._metadata
+
+    @metadata.setter
+    def metadata(self, metadata: Dict[str, Any]) -> None:
+        """Sets the metadata structure."""
+        self._metadata = metadata
+
+    def property_exists(self, prop: str) -> bool:
+        """Check if a given property is set.
+
+        Args:
+            prop: the property whose existence to check
+        """
+        return prop in self._metadata
+
+    def property(self, prop: str) -> Optional[Any]:
+        """Returns the value of a metadata property or None if it does not exist.
+
+        Args:
+            prop: the property to read
+        """
+        return self._metadata.get(prop, None)
+
+    def set_property(self, prop: str, value: Optional[Any]) -> None:
+        """Sets the value of a metadata property. For system properties, this function casts
+        the value to the proper type and throw an exception if this fails. If the value is
+        None, the property is removed.
+
+        Args:
+            prop: property to set
+            value: value to assign
+        """
+        if value is None:
+            self._metadata.pop(prop, None)
+        else:
+            if prop in METADATA_SYSTEM_FIELD_TYPES:
+                value = METADATA_SYSTEM_FIELD_TYPES[prop](value)
+            self._metadata[prop] = value
+
+    def clear_metadata(self) -> None:
+        """Remove all properties."""
+        self._metadata = {}
+
+
 # noinspection PyShadowingNames
-class LineCollection:
+class LineCollection(_MetadataMixin):
     """
     :py:class:`LineCollection` encapsulate a list of piecewise linear lines (or paths). Lines
     are implemented as 1D numpy arrays of complex numbers whose real and imaginary parts
@@ -86,17 +156,20 @@ class LineCollection:
 
     Finally, :py:class:`LineCollection` implements a number of operations such as geometrical
     transformation, cropping, merging, etc. (see member function documentation for details).
+
+    Args:
+        lines (LineCollectionLike): iterable of line (accepts the same input as
+            :func:`~LineCollection.append`).
+        metadata: if provided, used as layer metadata
     """
 
-    def __init__(self, lines: LineCollectionLike = ()):
-        """Create a LineCollection instance from an iterable of lines.
+    def __init__(
+        self, lines: LineCollectionLike = (), metadata: Optional[Dict[str, Any]] = None
+    ):
+        """Create a LineCollection instance from an iterable of lines."""
+        super().__init__(metadata)
 
-        Args:
-            lines (LineCollectionLike): iterable of line (accepts the same input as
-                :func:`~LineCollection.append`).
-        """
         self._lines: List[np.ndarray] = []
-
         self.extend(lines)
 
     @property
@@ -108,6 +181,20 @@ class LineCollection:
         """
         return self._lines
 
+    def clone(self, lines: LineCollectionLike = ()) -> "LineCollection":
+        """Creates a new :class:`LineCollection` with the same metadata.
+
+        If ``lines`` is provided, its content is added to the new :class:`LineCollection`
+        instance.
+
+        Args:
+            lines: path to add to the new :class:`LineCollection` instance
+
+        Returns:
+            the new :class:`LineCollection` instance
+        """
+        return LineCollection(lines=lines, metadata=self.metadata)
+
     def append(self, line: LineLike) -> None:
         """Append a single line.
 
@@ -117,9 +204,9 @@ class LineCollection:
         Args:
             line (LineLike): line to append
         """
-        if isinstance(line, LineString) or isinstance(line, LinearRing):
+        if isinstance(line, (LineString, LinearRing)):
             # noinspection PyTypeChecker
-            self._lines.append(np.array(line).view(dtype=complex).reshape(-1))
+            self._lines.append(np.array(line.coords).view(dtype=complex).reshape(-1))
         else:
             line = np.array(line, dtype=complex).reshape(-1)
             if len(line) > 1:
@@ -143,8 +230,10 @@ class LineCollection:
         if hasattr(lines, "geom_type") and lines.is_empty:  # type: ignore
             return
 
-        # sometimes, mls end up actually being ls
-        if isinstance(lines, LineString) or isinstance(lines, LinearRing):
+        # handle shapely objects
+        if isinstance(lines, MultiLineString):
+            lines = lines.geoms
+        elif isinstance(lines, (LineString, LinearRing)):
             lines = [lines]
 
         for line in lines:
@@ -329,7 +418,7 @@ class LineCollection:
                     line = np.flip(line)
                 if idx is None:
                     break
-                new_line = index.pop(idx)
+                new_line = cast(np.ndarray, index.pop(idx))
                 if reverse:
                     new_line = np.flip(new_line)
                 line = np.hstack([line, new_line])
@@ -349,10 +438,10 @@ class LineCollection:
             return None
         else:
             return (
-                min((line.real.min() for line in self._lines)),
-                min((line.imag.min() for line in self._lines)),
-                max((line.real.max() for line in self._lines)),
-                max((line.imag.max() for line in self._lines)),
+                float(min((line.real.min() for line in self._lines))),
+                float(min((line.imag.min() for line in self._lines))),
+                float(max((line.real.max() for line in self._lines))),
+                float(max((line.imag.max() for line in self._lines))),
             )
 
     def width(self) -> float:
@@ -363,8 +452,9 @@ class LineCollection:
         """
 
         if self._lines:
-            return max((line.real.max() for line in self._lines)) - min(
-                (line.real.min() for line in self._lines)
+            return float(
+                max((line.real.max() for line in self._lines))
+                - min((line.real.min() for line in self._lines))
             )
         else:
             return 0.0
@@ -376,8 +466,9 @@ class LineCollection:
             the width (ymax - ymin) or 0.0 if the LineCollection is empty
         """
         if self._lines:
-            return max((line.imag.max() for line in self._lines)) - min(
-                (line.imag.min() for line in self._lines)
+            return float(
+                max((line.imag.max() for line in self._lines))
+                - min((line.imag.min() for line in self._lines))
             )
         else:
             return 0.0
@@ -412,7 +503,7 @@ class LineCollection:
         starts = np.array([line[0] for line in self.lines[1:]])
         dists = np.abs(starts - ends)
         # noinspection PyTypeChecker
-        return np.sum(dists), np.mean(dists), np.median(dists)
+        return float(np.sum(dists)), float(np.mean(dists)), float(np.median(dists))
 
     def segment_count(self) -> int:
         """Returns the total number of segment across all lines.
@@ -423,7 +514,7 @@ class LineCollection:
         return sum(max(0, len(line) - 1) for line in self._lines)
 
 
-class Document:
+class Document(_MetadataMixin):
     """This class is the core data model of vpype and represent the data that is passed from
     one command to the other. At its core, a Document is a collection of layers identified
     by non-zero positive integers and each represented by a :py:class:`LineCollection`.
@@ -431,11 +522,17 @@ class Document:
     In addition, the Document class maintains a :py:attr:`page_size` attribute which describe
     the physical size of the document. This attribute is not strictly linked to the actual
     Document's content, but can be set based on it.
+
+    Args:
+        line_collection: if provided, used as layer 1
+        metadata: if provided, used as global metadata
+        page_size: if provided, used as page size
     """
 
     def __init__(
         self,
         line_collection: LineCollection = None,
+        metadata: Optional[Dict[str, Any]] = None,
         page_size: Optional[Tuple[float, float]] = None,
     ):
         """Create a Document, optionally providing a :py:class:`LayerCollection` for layer 1.
@@ -443,15 +540,38 @@ class Document:
         Args:
             line_collection: if provided, used as layer 1
         """
+        super().__init__(metadata)
+
         self._layers: Dict[int, LineCollection] = {}
-        self._page_size: Optional[Tuple[float, float]] = page_size
+
+        if page_size is not None:
+            self.page_size = page_size
 
         if line_collection:
             self.add(line_collection, 1)
 
-    def empty_copy(self) -> "Document":
-        """Create an empty copy of this document with the same page size."""
-        return Document(page_size=self.page_size)
+    def clone(self, keep_layers: bool = False) -> "Document":
+        """Create an empty copy of this document with the same metadata.
+
+        By default, the cloned document doest not contain any layer. If ``keep_layers`` is set
+        to true, empty layers with metadata will be created to match the source document's
+        layer.
+
+        Args:
+            keep_layers: if True, empty layers matching the source document's layers and
+                metadata
+
+        Returns:
+            the cloned document
+        """
+        doc = Document(metadata=self.metadata)
+        if keep_layers:
+            for layer_id in self.layers:
+                doc.layers[layer_id] = self.layers[layer_id].clone()
+        return doc
+
+    # backward compatibility
+    empty_copy = clone
 
     @property
     def layers(self) -> Dict[int, LineCollection]:
@@ -464,12 +584,12 @@ class Document:
     @property
     def page_size(self) -> Optional[Tuple[float, float]]:
         """Returns the page size or None if it hasn't been set."""
-        return self._page_size
+        return self.property(METADATA_FIELD_PAGE_SIZE)
 
     @page_size.setter
     def page_size(self, page_size=Optional[Tuple[float, float]]) -> None:
         """Sets the page size to a new value."""
-        self._page_size = page_size
+        self.set_property(METADATA_FIELD_PAGE_SIZE, page_size)
 
     def extend_page_size(self, page_size: Optional[Tuple[float, float]]) -> None:
         """Adjust the  page sized according to the following logic:
@@ -491,6 +611,35 @@ class Document:
             else:
                 self.page_size = page_size
 
+    @property
+    def sources(self) -> Set[pathlib.Path]:
+        return self.metadata.get(METADATA_FIELD_SOURCE_LIST, set())
+
+    @sources.setter
+    def sources(self, sources: Set[pathlib.Path]) -> None:
+        self.set_property(METADATA_FIELD_SOURCE_LIST, sources)
+
+    def add_to_sources(self, path) -> None:
+        """Add a path to the source list.
+
+        If ``path`` cannot be converted to a :class:`pathlib.Path` or the file doesn't exist,
+        it is ignored and not added to the source list.
+
+        Args:
+            path: file path
+        """
+        try:
+            path = pathlib.Path(path)
+            if path.exists():
+                self.sources |= {path}
+        except TypeError:
+            pass
+
+    def clear_layer_metadata(self) -> None:
+        """Clear all metadata from the document."""
+        for layer in self._layers.values():
+            layer.clear_metadata()
+
     def ids(self) -> Iterable[int]:
         """Returns the list of layer IDs"""
         return self._layers.keys()
@@ -510,10 +659,11 @@ class Document:
         """
         return (self._layers[lid] for lid in layer_ids if lid in self._layers)
 
-    def exists(self, layer_id: int) -> bool:
+    def exists(self, layer_id: Optional[int]) -> bool:
         """Test existence of a layer.
 
-        Note that existence of a layer does not necessarily imply that it isn't empty.
+        Note that existence of a layer does not necessarily imply that it isn't empty. Always
+        return False if ``layer_id`` is None.
 
         Args:
             layer_id: layer ID to test
@@ -521,6 +671,9 @@ class Document:
         Returns:
             True if the layer ID exists
         """
+        if layer_id is None:
+            return False
+
         return layer_id in self._layers
 
     def __getitem__(self, layer_id: int):
@@ -546,21 +699,89 @@ class Document:
             vid += 1
         return vid
 
-    def add(self, lc: LineCollection, layer_id: Union[None, int] = None) -> None:
+    def add(
+        self,
+        lines: LineCollectionLike,
+        layer_id: Union[None, int] = None,
+        with_metadata: bool = False,
+    ) -> None:
         """Add a the content of a :py:class:`LineCollection` to a given layer.
 
-        If the given layer is None, the input LineCollection is used to create a new layer
-        using the lowest available layer ID.
+        If the given layer ID is :data:`None`, a new layer with the lowest available layer ID
+        is created and initialized with the content of ``lc``.
+
+        The destination layer's metadata is unchanged, unless ``with_metadata`` is
+        :data:`True`.
+
+        Args:
+            lines: lines to add to the layer
+            layer_id: the destination layer ID, or :data:`None` for a new layer
+            with_metadata: copies/update the metadata to the destination layer (if any)
         """
         if layer_id is None:
             layer_id = 1
             while layer_id in self._layers:
                 layer_id += 1
 
-        if layer_id in self._layers:
-            self._layers[layer_id].extend(lc)
-        else:
-            self._layers[layer_id] = lc
+        if layer_id not in self._layers:
+            self._layers[layer_id] = LineCollection()
+
+        self._layers[layer_id].extend(lines)
+
+        if with_metadata and isinstance(lines, LineCollection):
+            self._layers[layer_id].metadata.update(lines.metadata)
+
+    def replace(
+        self, lines: LineCollectionLike, layer_id: int, *, with_metadata: bool = False
+    ) -> None:
+        """Replaces the content of a layer.
+
+        This method replaces the content of layer ``layer_id`` with ``lines``. The layer must
+        exist, otherwise a :class:`ValueError` exception is raised.
+
+        The destination layer retains its metadata, unless ``with_metadata`` is True, in which
+        case ``lines`` must have a ``metadata`` attribute.
+
+        Args:
+            lines: line data to assign to the layer
+            layer_id: layer ID of the layer whose content is to be replaced
+            with_metadata:
+        """
+
+        if layer_id not in self._layers:
+            raise ValueError(f"layer {layer_id} doesn't exist")
+
+        self._layers[layer_id] = self._layers[layer_id].clone(lines)
+
+        if with_metadata:
+            metadata = getattr(lines, "metadata", None)
+            if metadata is not None:
+                self._layers[layer_id].metadata = metadata.copy()
+
+    def swap_content(self, lid1: int, lid2: int) -> None:
+        """Swap the content of two layers.
+
+        This method swaps the content between two layers. Each layer retains its original
+        metadata.
+
+        A :class:`ValueError` is raised if either of the layer ID do not exist.
+
+        Args:
+            lid1: first layer
+            lid2: second layer
+        """
+
+        if lid1 not in self._layers:
+            raise ValueError(f"layer {lid1} does not exist")
+        if lid2 not in self._layers:
+            raise ValueError(f"layer {lid2} does not exist")
+
+        self._layers[lid1].metadata, self._layers[lid2].metadata = (
+            self._layers[lid2].metadata,
+            self._layers[lid1].metadata,
+        )
+
+        self._layers[lid1], self._layers[lid2] = self._layers[lid2], self._layers[lid1]
 
     def extend(self, doc: "Document") -> None:
         """Extend a Document with the content of another Document.
@@ -571,14 +792,28 @@ class Document:
 
         The :py:attr:`page_size` attribute is adjusted using :meth:`extend_page_size`.
 
+        The source's metadata is copied over to the destination, overriding any existing
+        property. Properties existing in the destination but not in the source are preserved.
+
         Args:
             doc: source Document
         """
 
+        # special treatment for page size and source list
         self.extend_page_size(doc.page_size)
+        for path in doc.sources:
+            self.add_to_sources(path)
+        self.metadata.update(
+            {
+                k: v
+                for k, v in doc.metadata.items()
+                if k not in {METADATA_FIELD_PAGE_SIZE, METADATA_FIELD_SOURCE_LIST}
+            }
+        )
 
         for layer_id, layer in doc.layers.items():
             self.add(layer, layer_id)
+            self.layers[layer_id].metadata.update(layer.metadata)
 
     def is_empty(self) -> bool:
         """Returns True if all layers are empty.
@@ -722,11 +957,3 @@ class Document:
             the total number of segments in the geometries
         """
         return sum(layer.segment_count() for layer in self._layers.values())
-
-
-class VectorData(Document):
-    """Deprecated, use Document."""
-
-    def __init__(self, *args, **kwargs):
-        super().__init__(*args, **kwargs)
-        logging.warning("!!! `vpype.VectorData` is deprecated, use `vpype.Document` instead.")
